@@ -1,7 +1,8 @@
 /*
     xattr - a python module for manipulating filesystem extended attributes
 
-    Copyright (C) 2002, 2003, 2006, 2008, 2012, 2013 Iustin Pop <iusty@k1024.org>
+    Copyright (C) 2002, 2003, 2006, 2008, 2012, 2013, 2015
+      Iustin Pop <iustin@k1024.org>
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -147,7 +148,13 @@ static int convert_obj(PyObject *myobj, target_t *tgt, int nofollow) {
         tgt->type = nofollow ? T_LINK : T_PATH;
         tgt->tmp = \
             PyUnicode_AsEncodedString(myobj,
-                                      Py_FileSystemDefaultEncoding, "strict");
+                                      Py_FileSystemDefaultEncoding,
+#ifdef IS_PY3K
+                                      "surrogateescape"
+#else
+                                      "strict"
+#endif
+                                      );
         if(tgt->tmp == NULL)
             return -1;
         tgt->name = PyBytes_AS_STRING(tgt->tmp);
@@ -269,7 +276,8 @@ pygetxattr(PyObject *self, PyObject *args)
     int nofollow = 0;
     char *attrname = NULL;
     char *buf;
-    ssize_t nalloc, nret;
+    ssize_t nalloc_s, nret;
+    size_t nalloc;
     PyObject *res;
 
     /* Parse the arguments */
@@ -277,36 +285,38 @@ pygetxattr(PyObject *self, PyObject *args)
         return NULL;
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     /* Find out the needed size of the buffer */
-    if((nalloc = _get_obj(&tgt, attrname, NULL, 0)) == -1) {
+    if((nalloc_s = _get_obj(&tgt, attrname, NULL, 0)) == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freetgt;
+        goto free_tgt;
     }
 
+    nalloc = (size_t) nalloc_s;
+
     /* Try to allocate the memory, using Python's allocator */
     if((buf = PyMem_Malloc(nalloc)) == NULL) {
         res = PyErr_NoMemory();
-        goto freetgt;
+        goto free_tgt;
     }
 
     /* Now retrieve the attribute value */
     if((nret = _get_obj(&tgt, attrname, buf, nalloc)) == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freebuf;
+        goto free_buf;
     }
 
     /* Create the string which will hold the result */
     res = PyBytes_FromStringAndSize(buf, nret);
 
- freebuf:
+ free_buf:
     /* Free the buffer, now it is no longer needed */
     PyMem_Free(buf);
- freetgt:
+ free_tgt:
     free_tgt(&tgt);
- freearg:
+ free_arg:
     PyMem_Free(attrname);
 
     /* Return the result */
@@ -346,7 +356,8 @@ xattr_get(PyObject *self, PyObject *args, PyObject *keywds)
     const char *fullname;
     char *buf;
     const char *ns = NULL;
-    ssize_t nalloc, nret;
+    ssize_t nalloc_s, nret;
+    size_t nalloc;
     PyObject *res;
     static char *kwlist[] = {"item", "name", "nofollow", "namespace", NULL};
 
@@ -356,43 +367,45 @@ xattr_get(PyObject *self, PyObject *args, PyObject *keywds)
         return NULL;
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     if(merge_ns(ns, attrname, &fullname, &namebuf) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_tgt;
     }
 
     /* Find out the needed size of the buffer */
-    if((nalloc = _get_obj(&tgt, fullname, NULL, 0)) == -1) {
+    if((nalloc_s = _get_obj(&tgt, fullname, NULL, 0)) == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freetgt;
+        goto free_name_buf;
     }
 
+    nalloc = (size_t) nalloc_s;
+
     /* Try to allocate the memory, using Python's allocator */
     if((buf = PyMem_Malloc(nalloc)) == NULL) {
         res = PyErr_NoMemory();
-        goto freenamebuf;
+        goto free_name_buf;
     }
 
     /* Now retrieve the attribute value */
     if((nret = _get_obj(&tgt, fullname, buf, nalloc)) == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freebuf;
+        goto free_buf;
     }
 
     /* Create the string which will hold the result */
     res = PyBytes_FromStringAndSize(buf, nret);
 
     /* Free the buffers, they are no longer needed */
- freebuf:
+ free_buf:
     PyMem_Free(buf);
- freenamebuf:
+ free_name_buf:
     PyMem_Free(namebuf);
- freetgt:
+ free_tgt:
     free_tgt(&tgt);
- freearg:
+ free_arg:
     PyMem_Free(attrname);
 
     /* Return the result */
@@ -443,9 +456,10 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
     PyObject *myarg, *res;
     int nofollow=0;
     const char *ns = NULL;
-    char *buf_list, *buf_val;
+    char *buf_list, *buf_val, *buf_val_tmp;
     const char *s;
-    ssize_t nalloc, nlist, nval;
+    ssize_t nalloc_s, nlist, nval_s;
+    size_t nalloc, nval;
     PyObject *mylist;
     target_t tgt;
     static char *kwlist[] = {"item", "nofollow", "namespace", NULL};
@@ -460,17 +474,24 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
     /* Compute first the list of attributes */
 
     /* Find out the needed size of the buffer for the attribute list */
-    nalloc = _list_obj(&tgt, NULL, 0);
+    nalloc_s = _list_obj(&tgt, NULL, 0);
 
-    if(nalloc == -1) {
+    if(nalloc_s == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freetgt;
+        goto free_tgt;
     }
 
+    if(nalloc_s == 0) {
+        res = PyList_New(0);
+        goto free_tgt;
+    }
+
+    nalloc = (size_t) nalloc_s;
+
     /* Try to allocate the memory, using Python's allocator */
     if((buf_list = PyMem_Malloc(nalloc)) == NULL) {
         res = PyErr_NoMemory();
-        goto freetgt;
+        goto free_tgt;
     }
 
     /* Now retrieve the list of attributes */
@@ -506,17 +527,27 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
         /* Now retrieve the attribute value */
         missing = 0;
         while(1) {
-            nval = _get_obj(&tgt, s, buf_val, nalloc);
+            nval_s = _get_obj(&tgt, s, buf_val, nalloc);
 
-            if(nval == -1) {
+            if(nval_s == -1) {
                 if(errno == ERANGE) {
-                    nval = _get_obj(&tgt, s, NULL, 0);
-                    if((buf_val = PyMem_Realloc(buf_val, nval)) == NULL) {
+                    ssize_t realloc_size_s = _get_obj(&tgt, s, NULL, 0);
+                    /* ERANGE + proper size should not fail, but it
+                       still can, so let's check first */
+                    if(realloc_size_s == -1) {
+                      res = PyErr_SetFromErrno(PyExc_IOError);
+                      Py_DECREF(mylist);
+                      goto free_buf_val;
+                    }
+                    size_t realloc_size = (size_t) realloc_size_s;
+                    if((buf_val_tmp = PyMem_Realloc(buf_val, realloc_size))
+                       == NULL) {
                         res = PyErr_NoMemory();
                         Py_DECREF(mylist);
-                        goto free_buf_list;
+                        goto free_buf_val;
                     }
-                    nalloc = nval;
+                    buf_val = buf_val_tmp;
+                    nalloc = realloc_size;
                     continue;
                 } else if(errno == ENODATA || errno == ENOATTR) {
                     /* this attribute has gone away since we queried
@@ -528,7 +559,9 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
                    don't know how to handle nicely, so we abort */
                 Py_DECREF(mylist);
                 res = PyErr_SetFromErrno(PyExc_IOError);
-                goto freebufval;
+                goto free_buf_val;
+            } else {
+              nval = (size_t) nval_s;
             }
             break;
         }
@@ -542,7 +575,7 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
         if (my_tuple == NULL) {
           Py_DECREF(mylist);
           res = NULL;
-          goto freebufval;
+          goto free_buf_val;
         }
         PyList_Append(mylist, my_tuple);
         Py_DECREF(my_tuple);
@@ -551,13 +584,13 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
     /* Successful exit */
     res = mylist;
 
- freebufval:
+ free_buf_val:
     PyMem_Free(buf_val);
 
  free_buf_list:
     PyMem_Free(buf_list);
 
- freetgt:
+ free_tgt:
     free_tgt(&tgt);
 
     /* Return the result */
@@ -593,18 +626,28 @@ pysetxattr(PyObject *self, PyObject *args)
     int nofollow = 0;
     char *attrname = NULL;
     char *buf = NULL;
-    int bufsize;
+    Py_ssize_t bufsize_s;
+    size_t bufsize;
     int nret;
     int flags = 0;
     target_t tgt;
 
     /* Parse the arguments */
     if (!PyArg_ParseTuple(args, "Oetet#|ii", &myarg, NULL, &attrname,
-                          NULL, &buf, &bufsize, &flags, &nofollow))
+                          NULL, &buf, &bufsize_s, &flags, &nofollow))
         return NULL;
+
+    if (bufsize_s < 0) {
+        PyErr_SetString(PyExc_ValueError,
+                        "negative value size?!");
+        res = NULL;
+        goto free_arg;
+    }
+    bufsize = (size_t) bufsize_s;
+
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     /* Set the attribute's value */
@@ -614,13 +657,13 @@ pysetxattr(PyObject *self, PyObject *args)
 
     if(nret == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freearg;
+        goto free_arg;
     }
 
     Py_INCREF(Py_None);
     res = Py_None;
 
- freearg:
+ free_arg:
     PyMem_Free(attrname);
     PyMem_Free(buf);
 
@@ -659,7 +702,8 @@ xattr_set(PyObject *self, PyObject *args, PyObject *keywds)
     int nofollow = 0;
     char *attrname = NULL;
     char *buf = NULL;
-    int bufsize;
+    Py_ssize_t bufsize_s;
+    size_t bufsize;
     int nret;
     int flags = 0;
     target_t tgt;
@@ -672,16 +716,25 @@ xattr_set(PyObject *self, PyObject *args, PyObject *keywds)
     /* Parse the arguments */
     if (!PyArg_ParseTupleAndKeywords(args, keywds, "Oetet#|ii" BYTES_CHAR,
                                      kwlist, &myarg, NULL, &attrname, NULL,
-                                     &buf, &bufsize, &flags, &nofollow, &ns))
+                                     &buf, &bufsize_s, &flags, &nofollow, &ns))
         return NULL;
+
+    if (bufsize_s < 0) {
+        PyErr_SetString(PyExc_ValueError,
+                        "negative value size?!");
+        res = NULL;
+        goto free_arg;
+    }
+    bufsize = (size_t) bufsize_s;
+
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     if(merge_ns(ns, attrname, &full_name, &newname) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     /* Set the attribute's value */
@@ -693,13 +746,13 @@ xattr_set(PyObject *self, PyObject *args, PyObject *keywds)
 
     if(nret == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freearg;
+        goto free_arg;
     }
 
     Py_INCREF(Py_None);
     res = Py_None;
 
- freearg:
+ free_arg:
     PyMem_Free(attrname);
     PyMem_Free(buf);
 
@@ -736,7 +789,7 @@ pyremovexattr(PyObject *self, PyObject *args)
 
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     /* Remove the attribute */
@@ -746,13 +799,13 @@ pyremovexattr(PyObject *self, PyObject *args)
 
     if(nret == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freearg;
+        goto free_arg;
     }
 
     Py_INCREF(Py_None);
     res = Py_None;
 
- freearg:
+ free_arg:
     PyMem_Free(attrname);
 
     /* Return the result */
@@ -798,12 +851,12 @@ xattr_remove(PyObject *self, PyObject *args, PyObject *keywds)
 
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     if(merge_ns(ns, attrname, &full_name, &name_buf) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     /* Remove the attribute */
@@ -815,13 +868,13 @@ xattr_remove(PyObject *self, PyObject *args, PyObject *keywds)
 
     if(nret == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freearg;
+        goto free_arg;
     }
 
     Py_INCREF(Py_None);
     res = Py_None;
 
- freearg:
+ free_arg:
     PyMem_Free(attrname);
 
     /* Return the result */
@@ -845,7 +898,8 @@ pylistxattr(PyObject *self, PyObject *args)
 {
     char *buf;
     int nofollow=0;
-    ssize_t nalloc, nret;
+    ssize_t nalloc_s, nret;
+    size_t nalloc;
     PyObject *myarg;
     PyObject *mylist;
     Py_ssize_t nattrs;
@@ -859,21 +913,28 @@ pylistxattr(PyObject *self, PyObject *args)
         return NULL;
 
     /* Find out the needed size of the buffer */
-    if((nalloc = _list_obj(&tgt, NULL, 0)) == -1) {
+    if((nalloc_s = _list_obj(&tgt, NULL, 0)) == -1) {
         mylist = PyErr_SetFromErrno(PyExc_IOError);
-        goto freetgt;
+        goto free_tgt;
     }
 
+    if(nalloc_s == 0) {
+        mylist = PyList_New(0);
+        goto free_tgt;
+    }
+
+    nalloc = (size_t) nalloc_s;
+
     /* Try to allocate the memory, using Python's allocator */
     if((buf = PyMem_Malloc(nalloc)) == NULL) {
         mylist = PyErr_NoMemory();
-        goto freetgt;
+        goto free_tgt;
     }
 
     /* Now retrieve the list of attributes */
     if((nret = _list_obj(&tgt, buf, nalloc)) == -1) {
         mylist = PyErr_SetFromErrno(PyExc_IOError);
-        goto freebuf;
+        goto free_buf;
     }
 
     /* Compute the number of attributes in the list */
@@ -884,7 +945,7 @@ pylistxattr(PyObject *self, PyObject *args)
     /* Create the list which will hold the result */
     mylist = PyList_New(nattrs);
     if(mylist == NULL)
-      goto freebuf;
+      goto free_buf;
 
     /* Create and insert the attributes as strings in the list */
     for(s = buf, nattrs = 0; s - buf < nret; s += strlen(s) + 1) {
@@ -892,17 +953,17 @@ pylistxattr(PyObject *self, PyObject *args)
         if(item == NULL) {
             Py_DECREF(mylist);
             mylist = NULL;
-            goto freebuf;
+            goto free_buf;
         }
         PyList_SET_ITEM(mylist, nattrs, item);
         nattrs++;
     }
 
- freebuf:
+ free_buf:
     /* Free the buffer, now it is no longer needed */
     PyMem_Free(buf);
 
- freetgt:
+ free_tgt:
     free_tgt(&tgt);
 
     /* Return the result */
@@ -940,7 +1001,8 @@ xattr_list(PyObject *self, PyObject *args, PyObject *keywds)
 {
     char *buf;
     int nofollow = 0;
-    ssize_t nalloc, nret;
+    ssize_t nalloc_s, nret;
+    size_t nalloc;
     PyObject *myarg;
     PyObject *res;
     const char *ns = NULL;
@@ -955,25 +1017,32 @@ xattr_list(PyObject *self, PyObject *args, PyObject *keywds)
         return NULL;
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     /* Find out the needed size of the buffer */
-    if((nalloc = _list_obj(&tgt, NULL, 0)) == -1) {
+    if((nalloc_s = _list_obj(&tgt, NULL, 0)) == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freetgt;
+        goto free_tgt;
     }
 
+    if(nalloc_s == 0) {
+        res = PyList_New(0);
+        goto free_tgt;
+    }
+
+    nalloc = (size_t) nalloc_s;
+
     /* Try to allocate the memory, using Python's allocator */
     if((buf = PyMem_Malloc(nalloc)) == NULL) {
         res = PyErr_NoMemory();
-        goto freetgt;
+        goto free_tgt;
     }
 
     /* Now retrieve the list of attributes */
     if((nret = _list_obj(&tgt, buf, nalloc)) == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freebuf;
+        goto free_buf;
     }
 
     /* Compute the number of attributes in the list */
@@ -984,7 +1053,7 @@ xattr_list(PyObject *self, PyObject *args, PyObject *keywds)
     /* Create the list which will hold the result */
     res = PyList_New(nattrs);
     if(res == NULL)
-        goto freebuf;
+        goto free_buf;
 
     /* Create and insert the attributes as strings in the list */
     for(s = buf, nattrs = 0; s - buf < nret; s += strlen(s) + 1) {
@@ -994,20 +1063,20 @@ xattr_list(PyObject *self, PyObject *args, PyObject *keywds)
             if(item == NULL) {
                 Py_DECREF(res);
                 res = NULL;
-                goto freebuf;
+                goto free_buf;
             }
             PyList_SET_ITEM(res, nattrs, item);
             nattrs++;
         }
     }
 
- freebuf:
+ free_buf:
     /* Free the buffer, now it is no longer needed */
     PyMem_Free(buf);
 
- freetgt:
+ free_tgt:
     free_tgt(&tgt);
- freearg:
+ free_arg:
 
     /* Return the result */
     return res;
